#!/usr/bin/python

import datetime
import logging
import os, os.path
import random
import re
import string
import sys
import yaml
import urllib2
import uec_instances

from twisted.internet import reactor, task
from twisted.python import usage
from zipfile import ZipFile, is_zipfile
from shutil import rmtree
from collections import defaultdict, namedtuple

import uec_utils

class Options(usage.Options):
    optFlags = [
        ["write_log", 'w', "Write log to file" ],
        ["max_instances", 'm', "Run the maximum number of instances" ],
    ]
    optParameters = [
        ["log", "l", "info", 
         "log level wanted (debug, info, warning, error, critical"],
        ["start_stop", "s", None, 
         "start and stop instances over a certain period of time (x minutes)"],
    ]

euca_create_volume    = 'euca-create-volume'
euca_delete_volume    = 'euca-delete-volume'
euca_attach_volume    = 'euca-attach-volume'
euca_detach_volume    = 'euca-detach-volume'
euca_create_snapshot  = 'euca-create-snapshot'
euca_delete_snapshot  = 'euca-delete-snapshot'
euca_describe_volsnap = 'euca-describe-volumes && euca-describe-snapshots'

volumes    = []
snapshots  = []
admin_user = None

class Volume(object):   
    def __init__(self, cluster, user):
        self.cluster       = cluster
        self.size          = None
        self.snapshot      = None
        self.user          = user
        self.id            = None
        self._state        = None
        self.instance      = None
        self.device        = None

    def __repr__(self):
        return "<Volume: %s - %s - %s - %s>" % (self.id, self.user,
                                                self.cluster, self.state)
        
    @property
    def logger(self):
        if self.id:
            return logging.getLogger("VOLUME %s" % (self.id))
        else:
            return logging.getLogger()        
        
    @property
    def state(self):
        return self._state
    @state.setter
    def state(self, value):
        self.logger.debug("State: %s" % (value))
        # The state has changed
        if self._state != value:
            self.logger.debug("New state: %s => %s" % (self._state, value))
            self._state = value        
  
    def getProcessOutput(self, executable, args=(), timeout=60):
        self.logger.debug("Executing: %s - %s" % (executable, args))
        return self.user.getProcessOutput(executable, args, timeout)            
            
    def create(self, **kw):
        if kw.has_key('snapshot'):
            self.snapshot = kw['snapshot']
            self.logger.info("Creating volume from %s" %self.snapshot)
            output = self.getProcessOutput(euca_create_volume,
                                           args= ['--snapshot', self.snapshot,
                                                  '-z', self.cluster])
            output.addCallbacks(self.created, self.errCreated)
        elif kw.has_key('size'):
            self.size = kw['size']
            self.logger.info("Creating volume with size = %s GB" %self.size)
            output = self.getProcessOutput(euca_create_volume,
                                           args= ['-s', self.size,
                                                  '-z', self.cluster])
            output.addCallbacks(self.created, self.errCreated)

    def created(self, output):
        self.logger.debug("Volume creation output: %s" % (output))
        created = False
        for l in output.split("\n"):
            if l.startswith('VOLUME'):
                self.id = l.split()[1]
                if self.id == "Null":
                    # volume creation failed
                    continue
                self.logger.info("Created in %s" %self.cluster)
                created = True
        if not created:
            self.errCreated(output)
     
    def errCreated(self, output):
        self.logger.error("Volume creation failed: %s" % (output))            
            
    def delete(self):
        self.logger.info("Deleting volume")
        output = self.getProcessOutput(euca_delete_volume, args = [self.id])
        output.addCallback(self.deleted)
    
    def deleted(self, output):
        self.logger.info("Deleted")
        self.state = "deleted"
        self.logger.debug("delete output: %s" % (output))

    def attach(self, instance, device='/dev/sdc'):
        self.logger.info("Attaching volume to instance %s as %s" %( instance.id,
                                                                    device))
        self.instance = instance
        self.device   = device
        output = self.getProcessOutput(euca_attach_volume,
                                       args= ['-i', instance.id,
                                              '-d', device, self.id])
        output.addCallbacks(self.attached, self.errAttached)        

    def errAttached(self, output):
        self.logger.warning("Failed to attach volume to %s : %s" %
                            (self.instance.id, output))
        
    def attached(self, output):
        self.logger.debug("Volume attach output: %s" % (output))
        attached = False
        for l in output.split("\n"):
            if l.startswith('VOLUME'):
                self.id = l.split()[1]
                self.logger.info("Attached (%s/%s)" % (self.instance.id,
                                                       self.device))
                attached = True
        if not attached:
            self.errAttached(output) 
    
    def detach(self):
        self.logger.info("Detaching volume")
        self.instance = None
        self.device   = None        
        output = self.getProcessOutput(euca_detach_volume, args = [self.id])
        output.addCallback(self.detached)
    
    def detached(self, output):
        self.logger.info("Detached")
        self.state = "detached"
        self.logger.debug("detach output: %s" % (output))      

class Snapshot(object):    
    def __init__(self, volume, user):
        self.volume        = volume
        self.user          = user
        self.id            = None
        self._state        = None

    def __repr__(self):
        return "<Snapshot: %s - %s - %s>" % (self.id, self.user, self.state)
        
    @property
    def logger(self):
        if self.id:
            return logging.getLogger("SNAPSHOT %s" % (self.id))
        else:
            return logging.getLogger()        
        
    @property
    def state(self):
        return self._state
    @state.setter
    def state(self, value):
        self.logger.debug("State: %s" % (value))
        # The state has changed
        if self._state != value:
            self.logger.debug("New state: %s => %s" % (self._state, value))
            self._state = value        
  
    def getProcessOutput(self, executable, args=(), timeout=60):
        self.logger.debug("Executing: %s - %s" % (executable, args))
        return self.user.getProcessOutput(executable, args, timeout)            
            
    def create(self):
        self.logger.info("Creating snapshot")
        output = self.getProcessOutput(euca_create_snapshot,
                                       args = [self.volume.id])
        output.addCallbacks(self.created, self.errCreated)

    def created(self, output):
        self.logger.debug("Snapshot creation output: %s" % (output))
        created = False
        for l in output.split("\n"):
            if l.startswith('SNAPSHOT'):
                self.id = l.split()[1]
                if self.id == "Null":
                    # snapshot creation failed
                    continue
                self.logger.info("Created")
                created = True
        if not created:
            self.errCreated(output)
     
    def errCreated(self, output):
        self.logger.error("Snapshot creation failed: %s" % (output))            
            
    def delete(self):
        self.logger.info("Deleting snapshot")
        output = self.getProcessOutput(euca_delete_snapshot, args = [self.id])
        output.addCallback(self.deleted)
    
    def deleted(self, output):
        self.logger.info("Deleted")
        self.state = "deleted"
        self.logger.debug("delete output: %s" % (output))
  
def checkVolumeSnapState(volumes, snapshots, admin_user):
    if not admin_user.ready:
        reactor.callLater(16, checkVolumeSnapState, volumes, snapshots,
                          admin_user)
        return
    logging.debug("Checking volumes/snapshots state")
    output = admin_user.getProcessOutput(euca_describe_volsnap)
    output.addCallback(VolumeSnapState, volumes, snapshots, admin_user)
    
def VolumeSnapState(output, volumes, snapshots, admin_user):
    logging.debug("describe volumes and snapshots: %s", output)
    for l in output.split("\n"):
        if l.startswith('VOLUME'):
            try:
                v_id = l.split()[1]
                v = filter(lambda v: v.id == v_id, volumes)[0]
            except IndexError:
                #logging.debug("Unknown volume %s - skipping" % (v_id))
                continue
            state =  l.split("\t")[5]
            v.state = state
        if l.startswith('SNAPSHOT'):
            try:
                s_id = l.split()[1]
                s = filter(lambda s: s.id == s_id, snapshots)[0]
            except IndexError:
                #logging.debug("Unknown snapshot %s - skipping" % (s_id))
                continue
            state =  l.split("\t")[3]
            s.state = state
    reactor.callLater(random.randint(10, 20) + random.random(),
                          checkVolumeSnapState, volumes, snapshots, admin_user)      

def FindBiggestInstance(instances):   
    logging.debug("Look for a High-CPU instance to attach a volume")    
    global volumes
    global admin_user 
    if len(instances) == len(filter(lambda i: i.test_state in \
                                    ( "success", "failed", "lost" ), 
                                    instances)):  
        big = filter(lambda j: j.test_state in 'success' and
        uec_instances.Instance.ZONE_CPU_DICT[j.vmtype].cpu == max( [
        uec_instances.Instance.ZONE_CPU_DICT[i.vmtype].cpu for i in instances
        ] ) and
        uec_instances.Instance.ZONE_CPU_DICT[j.vmtype].ram == max( [
        uec_instances.Instance.ZONE_CPU_DICT[i.vmtype].ram for i in instances
        ] ) and
        uec_instances.Instance.ZONE_CPU_DICT[j.vmtype].disk == max( [
        uec_instances.Instance.ZONE_CPU_DICT[i.vmtype].disk for i in instances
        ] ), instances)[0]
        logging.info("Found a High-CPU instance : %s" %big)
        v = Volume(cluster = big.cluster, user = admin_user)
        v.create(size = "1")
        volumes.append(v)
        reactor.callLater(30, AttachVolumetoBiggestInstance, big, v)
    else:
        reactor.callLater(random.randint(10, 15) + random.random(),
                  FindBiggestInstance, instances)    
   
def AttachVolumetoBiggestInstance(big_i, vol):
    if vol.state not in 'available':
        reactor.callLater(random.randint(5, 10) + random.random(),
                          AttachVolumetoBiggestInstance, big_i, vol)
        return
    big_i.test_state = 'being-tested'
    logging.info("Attaching %s to %s" %( vol, big_i.id ))    
    vol.attach(big_i)
    reactor.callLater(random.randint(10, 15) + random.random(),
                      MountVolumetoBiggestInstance, vol)

def MountVolumetoBiggestInstance(vol):
    if vol.state not in 'in-use':
        reactor.callLater(random.randint(5, 10) + random.random(),
                          MountVolumetoBiggestInstance, vol)
        return
    logging.info("Mounting %s as /ebs in %s" %( vol, vol.instance.id )) 
    cmd  = """sudo perl -e '@a=`lshw -C disk -short 2>&1`;"""
    cmd += """@a=pop(@a)=~/(\/dev\/.*?)\s+/; """
    cmd += """system("parted -s @a mktable msdos"""
    cmd += """&& parted -s -- @a mkpart primary ext3 0 -1 """
    cmd += """&& mke2fs @{a}1 && mkdir -p /ebs && mount @{a}1 /ebs")'""" 
    args = ['-o', 'UserKnownHostsFile=/dev/null',
            '-o', 'StrictHostKeyChecking=no',
            '-o', 'ConnectTimeout=5',
            '-o', 'Batchmode=yes',
            '-i', vol.instance.private_key,
            "ubuntu@%s" % (vol.instance.pub_ip)]
    logging.debug("Ssh args: %s" % (args+[cmd]))
    output = uec_utils.getProcessOutput('ssh', args+[cmd], errortoo=True)    
    output.addCallback(MountedBiggestInstance, vol, args)    
    
def MountedBiggestInstance(output, vol, args):
    logging.debug("Volume mount output: %s" % (output))  
    logging.info("Installing debootstrap in %s" %vol.instance.id) 
    cmd  = "sudo apt-get install --force-yes --yes debootstrap"
    logging.debug("Ssh args: %s" % (args+[cmd]))
    output = uec_utils.getProcessOutput('ssh', args+[cmd], errortoo=True)    
    output.addCallback(InstalldebootstrapBiggestInstance, vol, args)   
    
def InstalldebootstrapBiggestInstance(output, vol, args):
    logging.debug("debootstrap installation output: %s" % (output))  
    logging.info("Debootstrap lucid in /ebs for %s" %vol.instance.id) 
    cmd  = "sudo debootstrap lucid /ebs"
    logging.debug("Ssh args: %s" % (args+[cmd]))
    output = uec_utils.getProcessOutput('ssh', args+[cmd], errortoo=True)    
    output.addCallback(ChrootEBSBiggestInstance, vol, args)     

def ChrootEBSBiggestInstance(output, vol, args):    
    if not re.findall(r"Base system installed successfully", output, re.S):
        logging.error("debootstrap installation failed")
        reactor.stop()
    logging.info("Prepare chroot environnement in %s" %vol.instance.id) 
    cmd  = "sudo mount proc /ebs/proc -t proc && "
    cmd += "sudo mount sysfs /ebs/sys -t sysfs && "
    cmd += "sudo mount -o bind /dev /ebs/dev && "
    cmd += "sudo chroot /ebs apt-get install --force-yes --yes language-pack-en"
    logging.debug("Ssh args: %s" % (args+[cmd]))
    output = uec_utils.getProcessOutput('ssh', args+[cmd], errortoo=True)    
    output.addCallback(InstallServiceBiggestInstance, vol, args)   
  
def InstallServiceBiggestInstance(output, vol, args):  
    logging.debug("chroot env. installation output: %s" % (output)) 
    logging.info("Install service in chrooted env. in %s" %vol.instance.id) 
    cmd = "sudo chroot /ebs apt-get install --force-yes --yes apache2"
    logging.debug("Ssh args: %s" % (args+[cmd]))
    output = uec_utils.getProcessOutput('ssh', args+[cmd], errortoo=True)    
    output.addCallback(ServiceRunningBiggestInstance, vol, args)      
    
def ServiceRunningBiggestInstance(output, vol, args):
    logging.debug("service installation output: %s" % (output))  
    try:
        page = urllib2.urlopen('http://%s' %vol.instance.pub_ip).read()
    except IOError, e:
        if hasattr(e, 'reason'):
            logging.error(e.reason)
        elif hasattr(e, 'code'):
            logging.error(e.code)
        reactor.stop()
    else:
        if not (re.search(r"It works!", page)):
            logging.error('http://%s : "It works!" not found'
                          %vol.instance.pub_ip)
            reactor.stop()
        else:
            logging.info("Service running in %s" %vol.instance.id)   
            logging.info("Stop service chroot environnement in %s" %vol.instance.id)           
            cmd  = "sudo umount /ebs/sys && "
            cmd += "sudo umount /ebs/proc && "
            cmd += "sudo umount /ebs/dev && "
            cmd += "sudo chroot /ebs /etc/init.d/apache2 stop && "
            cmd += "sudo umount -lf /ebs" 
            logging.debug("Ssh args: %s" % (args+[cmd]))
            output = uec_utils.getProcessOutput('ssh', args+[cmd],
                                                errortoo=True)
            output.addCallback(CreateSnapshotfromInuseVolume, vol) 

def CreateSnapshotfromInuseVolume(output, vol):            
    global snapshots    
    logging.debug("service stop output: %s" % (output))   
    s = Snapshot(volume = vol, user = admin_user)
    s.create()
    snapshots.append(s)
    reactor.callLater(30, CreateNewVolumefromSnapshot, s, vol)
    
def CreateNewVolumefromSnapshot(snap, vol):
    if snap.state not in 'completed':
        reactor.callLater(random.randint(5, 10) + random.random(),
                          CreateNewVolumefromSnapshot, snap, vol)
        return
    vol.instance.test_state = "success"
    vol.detach()
    new_vol = Volume(cluster = vol.cluster, user = admin_user)
    new_vol.create(snapshot = snap.id)
    volumes.append(new_vol)
    reactor.callLater(5, WaitforNewVolume, vol, new_vol)

def WaitforNewVolume(vol, new_vol):
    if vol.state not in 'available' or new_vol.state not in 'available':
        reactor.callLater(random.randint(5, 10) + random.random(),
                          WaitforNewVolume, vol, new_vol)
        return
    logging.info("Test run completed")
    reactor.stop()

def main():
    LEVELS = {'debug': logging.DEBUG,
               'info': logging.INFO,
               'warning': logging.WARNING,
               'error': logging.ERROR,
               'critical': logging.CRITICAL}
    
    global volumes
    global snapshots
    global admin_user
    opts = Options()
    uec_instances.opts = opts
   
    try:
        opts.parseOptions() # When given no argument, parses sys.argv[1:]
    except usage.UsageError, errortext:
        print '%s: %s' % (sys.argv[0], errortext)
        print '%s: Try --help for usage details.' % (sys.argv[0])
        return 1 

    # Verify that script is run as root
    if os.getuid():
        sys.stderr.write(
            "This script needs superuser permissions to run correctly\n")
        return 1        
        
    level = LEVELS.get(opts['log'], logging.NOTSET)
   
    if opts['write_log']:
        suffix = "service"
        LOG_FILENAME = "/var/log/uec_ebs.%s.log" %suffix
        handler = logging.handlers.RotatingFileHandler(LOG_FILENAME,
                                                       backupCount=2)
        logging.getLogger().addHandler(handler)
        handler.setFormatter(
            logging.Formatter("[%(asctime)s] %(levelname)s %(message)s"))
        logging.getLogger().setLevel(level)
        handler.doRollover() 
    else:
        logging.basicConfig(level=level)
    logging.info("log level set to " + opts['log']) 
    instances  = [] 
    volumes    = []
    snapshots  = []
    admin_user = uec_instances.User() # Create admin user
    
    #Overload some imported stuff
    uec_instances.TerminaTestedInstances = FindBiggestInstance
    uec_instances.admin_user             = admin_user
    
    reactor.callWhenRunning(admin_user.setup) # setup admin user  
    reactor.callWhenRunning(uec_instances.checkRessources, instances,
                            admin_user, first_call = True)
    reactor.callWhenRunning(uec_instances.checkInstancesState, instances,
                            admin_user)
    reactor.callWhenRunning(checkVolumeSnapState, volumes, snapshots,
                            admin_user)
    uec_instances.printStats(instances)    
    reactor.run()
    uec_instances.cleanUp(instances)
    [ v.delete() for v in volumes ]
    [ s.delete() for s in snapshots ]
    uec_instances.printStats(instances, once=True)
    if len(filter(lambda i: i.test_state == 'success', instances)) != \
       len(instances):
        return 1
    else:
        return 0
    
if __name__ == "__main__":
    sys.exit(main())     
    